/* types.h
 * All the defines, and structures and Kexis uses.
 * Copyright (C) 2000 Wayde Milas (wmilas@rarcoa.com)
 * 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* defines
 */

#define VERSION_MAJOR 						0
#define VERSION_MINOR 						1
#define VERSION_SUB_MINOR 				4

#define DEBUG 										0
#define PRED_DEBUG 								0
#define DEBUG_LEEWAY 							100
#define DEBUG_POS 								50

#define COMPRESS      						0
#define DECOMPRESS    						1

#define MID 											0
#define AVE 											1	

#define		STANDARD_ENCODER 				3
#define   RICE_SINGLE_K						1
#define   RICE_DUAL_K							2
#define		RICE_DUAL_K_BLOCK				3
#define   MAX_STANDARD_ENCODER    3

#define		STANDARD_PREDICTOR 			3
#define		ZEROO_PREDICTOR					1
#define   FIRSTO_PREDICTOR        2
#define   SECONDO_PREDICTOR				3
#define		THIRDO_PREDICTOR				4
#define   L_HISTORY_PREDICTOR			5
#define   MAX_STANDARD_PREDICTOR	5

#define		MAX_M										100

#define		MAX_KHISTORY						256
#define   DEFAULT_KHISTORY				32

#define		INITIAL_KEXIS_HEADER		1
#define   CURRENT_KEXIS_HEADER		1

#define   UNITS										200000

#define		INITIAL_K								8
/* Structs
 */

typedef struct {
  char            ckID[4];
  unsigned long   ckSize;
  char            wave_ckID[4];
  char            fmt_ckID[4];
  unsigned long   fmt_ckSize;
  unsigned short  formatTag;
  unsigned short  nChannels;
  unsigned long   nSamplesPerSec;
  unsigned long   nAvgBytesPerSec;
  unsigned short  nBlockAlign;
  unsigned short  nBitsPerSample;
  char            data_ckID[4];
  unsigned long   data_ckSize;
} WAVHEADER;

// Dont change this struct until we update parse_kexis_header
typedef struct {
	char						kexisID[4];
	unsigned short	headerVersion;
	unsigned short	encoderVersion;
	unsigned short	predictorVersion;
	unsigned short  kHistorySize;
	unsigned long		dataLength;
	unsigned short	pad0;
	unsigned short  pad1;
	unsigned short  pad2;
	unsigned short  pad3;
} KEXISHEADER;

typedef struct {
  int             mode;
  int             progress;
	int							verbose;
  int             displayHeader;
	int							dec_stdout;
	int							kHistorySize;
	unsigned short	predictorVersion;
	unsigned short  encoderVersion;
  FILE            *inFileStream;
  FILE            *outFileStream;
  char            *inFileName;
  char            *outFileName;
	time_t					time;
	char						errorString[256];
} OPTIONSTRUCT;

typedef struct {
	int							mid_I[4];
	int							ave_I[4];
	int							midM;
	int							aveM;
	int							diffAve;
	int							diffMid;
	int							lowDiffAve;
	int							highDiffAve;
	double					sumDiffAve;
	double					countDiffAve;
	int     	      lowDiffMid;
	int      	  	  highDiffMid;
	double    		  sumDiffMid;
	double      	  countDiffMid;

	int             mid_K[MAX_KHISTORY];
	int							ave_K[MAX_KHISTORY];
	int							kBlockPointer;
	int             midLowK;
	int							aveLowK;
	int             midHighK;
	int							aveHighK;
	double          midSumK;
	double					aveSumK;
	double          midCountK;
	double					aveCountK;
} PREDICTORTABLE;

typedef struct {
  short           *data;
  long            dataPosition; //position in the block that data is stacked 
  															//to currently.
  PREDICTORTABLE  predictor;
  int             bitCounter;  //Keeps track of the halfway assembled word when
															 //encoding.
  long            bitBucket;   //used to hold partial words between bit pushes
	double					countK;
	unsigned long   blockPointer;
	unsigned long		subBlockPointer;
	unsigned long		sumTotalSize;
	unsigned long		debugBlockCount;
} KEXISBLOCKSTRUCT;

typedef struct {
  short           *data;
  unsigned long   size;
  unsigned long   pcmStreamLength;
} PCMBLOCKSTRUCT;
